
/* brs index lists - server */ 

/*
 * MODIFICATION HISTORY:
 *
 * xx/92  Leman Chung   Created.
 *
 * 8/93 Tom Unger       Added logic to keep connection to client open.
 *			Fork auxiliary server processes as needed.
 *			Added request number to requests and responces and
 *			appended EOT character to end of responce.
 *
 * 11/93 Tom Unger	Put all lists in one big table.  Only open lists
 *			as needed.  Made program more "unix daemon" like.
 *			instrumented server to generate logs fro performance
 *			enhancements.
 *
 * COMPILING
 *
 *	The signal calls are BSD standard.  This program must be linked with
 *	AIX's BDS signal compatibility library 'bsd':
 *
 *	cc -g -o brs_test3 brs_test3.c -lbsd
 *
 *	Or, USE THE MAKEFILE.
 *
 *
 * CONNECTED CLIENTS:
 *
 *	Originally, the server listened for connections, read a request,
 *	sent the responce, and closed the connection.  To eliminate the 
 *	overhead of creating and removing a connection for every request
 *	I've changed the code so that the client and server keep a 
 *	connection open between them.  
 *
 *	In order that the client can match responce with it's requests
 *	I added a request number that the client sends at the beginning
 *	of all requests.  The server places the same request number in
 *	the reply.  
 *
 *	The EOT character is sent at the end of the reply to mark the
 *	end of the reply.  Originally the client know when to stop reading
 *	because the connection got closed.  Since the connection stays
 *	open some marker in the data stream is needed.
 *
 *	The client sockes are set to NODELAY to expidite TCP transmission
 *	and improve the turn around time.
 *
 *
 * MULTIPLE PROCESSES:
 *
 *	Now the server keeps a connection with all active clients.  The
 *	exact number of clients is difficult to determine, but could be
 *	several hundred.  One process servicing several clients may lead
 *	to slow responce while the server process is handleing one client.
 *	Further, the server number of clients a server can handle is 
 *	limited by how many open files it may have.  (which on AIX is
 *	quite large:  2000).  Yet, we don't want to fork a single process 
 *	for each client - that would lead to several hundred processes.
 *	So, the server is written to handle up to some number, N, of
 *	clients.  When the number of clients surpasses N, the main
 *	server forks a child process.  The main server closes all client
 *	connections and continues to accept new connections.  The child
 *	process closes the server port and continues to service the
 *	active client connections.  When all the active clients close their
 *	connections the child process exits.  Meanwhile, the main process
 *	continued to accept and service new clients and may have created
 *	additional child processes if enough new connections came it.
 *	
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <sys/file.h>
#include <netdb.h>
#include <sys/stat.h>
#include <ctype.h>
#ifndef SUN
#include <sys/select.h>
#include <unistd.h>
#endif
#include <signal.h>
#include <sys/time.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/wait.h>

#include "brs_logger.h"

#ifdef SUN
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 1
#endif
#endif


/* The port we use.  Can be overridden with command line. */
#define SERV_TCP_PORT   3050

/* Default number of clients per process,  Can be overridden with command
 * line. */
#define DEF_CLIENT_LIMIT	1

/* Character sent at end of list signal the end of the list has been
 * reached. */
#define EOT	((char)0x04)

/* Extension added to list file name to the line index file name. */
#define LINE_INDEX_EXT	"_idx"

#define HDR_MAX         250
#define LINEMAX         1024
#define MAXPATH		250
#define Timeout         5
#define RS              0x1e

#define EMPTY_SOCKET	-1

#define SUCCESS		0
#define FAIL		1


/*
 * Return codes from the request processing routines str_extract and 
 * pul_data.
 */
#define REQUEST_SUCCESS 	0
#define REQUEST_ERROR 		-1
#define REQUEST_CLIENT_GONE	1



/*
 * TYPE DEFINITIONS.
 */

/*
 * Some lists have several different names.  We use a table of
 * aliases to convert from the possible names to a standard name.
 */
typedef struct {
    char	*alias;		/* Name supplied by client. */
    char	*name;		/* Gets converted to this name. */
} ListAlias;

/*
 * "list" list entry.
 * Each entry is indexed by a db_name and list_name.  
 *
 * Each entry either contains information for that list or a pointer
 * to another entry for a list that has exactly the same information.
 * This _could_ be implemented as a union, but I don't know how to 
 * make an initializer for a union.  So I include all the structures
 * but only one set or the other are used. 
 *
 * 'f_name' and 'hdr_name' must be supplied.  'lidx_name' need not.  If
 * it is not supplied then it will be generated by appending a 
 * standard extension to 'f_name'.
 */
typedef struct {

    /* Database name and list name. */
    char	*db_name;	/* DB name, e.g. "LCAT". */
    char	*list_name;	/* List name, e.g. "Subjects".*/
    int		etype;		/* Entry type. */

    /* File Information. */
    char	*f_name;	/* List file name. */
    FILE	*fp;		/* Pointer to list file. */
    char	*lidx_name;	/* Line index file name. */
    FILE	*lip;		/* Pointer to Index file. */
    char	*hdr_name;	/* Header File name. */
    FILE	*hp;		/* Pointer to header file. */
    long	file_lines;	/* Count of lines in file. */
    char	*hdr;		/* Pointer to header. */
    int		req_tblc;	/* another flag. */
    int		blk_to_hyphen;	/* some sort of flag. */
                                /* blk_to_hyphen = 1, drop hyphen later */
                                /* blk_to_hyphen = 2, keep hyphen later */


    /* Or a pointer to other list that has same info. */
    char	*rf_db_name;	/* Same as this DB. */
    char	*rf_list_name;	/* Same as this list. */
 } LList;

#define ETYPE_FILE		1
#define ETYPE_REF		2
#define FILES_PER_LIST		2	/* Number of files kept open for each
					   list. */



/*
 * Client record.  One per connectd client.
 */
typedef struct {
    int			socket;		/* Socket they are on. */
    unsigned int	id;		/* Their ID. */
    int			requests;	/* Number of requests made. */
    int			log_me;		/* TRUE if disconnect should be 
					   logged.*/
}  Client;





/*
 * GLOBAL VARIABLES.
 */
int	ClientLimit;		/* Clients per process. */
int	Debug;			/* Generate debugging output when non-zero.*/
char	*ProgName;		/* Name by which this program was invoked. */
int     ChildNo;		/* Number of child. */
int	ServerPort;		/* Port server listens on. */
Tick	SelectTime;		/* Time last select call returned. */
Tick	RequestTime;
LookupStats	LStats;		

/* This global variable is set to the requesting client ID before calling
 * str_extract().  If an error occurs while processing the request
 * the client ID in 'RequestingClientID' is included in the log message. */
unsigned int	RequestingClientID;

/* This global variable is set to the return value of LogProcCreate
 * which returns a boolean indicating if the process exit should 
 * be logged.  (Because if we log the creation we want to log the exit.)
 * Also, if an error occurs during any request, 'LogThisProcExit' is
 * set so we will haveee the record of when the proc exited. */
int	LogThisProcExit;	



/* The List Lookup function sets these global variables to indicate
 * which list to use. */
FILE 		*list, *list_idx, *list_req;
int  		req_tblc, blk_to_hyphen;
long 		file_lines;
char 		*hdr_rec_ptr;






/*
 * Forward procedure definitions. 
 */
void 		ProcessRequests (int sockfd);
void		daemon_init (void);
int 		str_extract(int sockfd);
int 		readline(char *ptr, char *output_str, int maxline);
long 		line_cnt(FILE *file_ptr);
void 		close_files(void); 
void 		file_check(char *db_name, char *list_name);
long 		line_search(long offset);
static void    	to_lower();

void		FListPrint (void);
void		FListTest (void);
int		FListCount (void);
LList *		FListFind (char *db_name, char *list_name);
int		FListOpen (LList *l);
void		FListExercise (void);
void 		TestLogger (void);
int		ErrorReplay (int sockfd, int req_number, char *msg);


/* 
 * TABLES.
 */

ListAlias list_alias[] = {
    {"Table of Contents", "Tbl-Contents"}, 
    {"Index of Subjects", "Subjects"}, 
    {"Newspapers", "Journal Titles"}, 
    {"Newspaper Titles", "Journal Titles"}, 
    {"Publisher Full Names", "Publishers"}, 
    {NULL, NULL}
};


LList	list_files[] = {

    {"ERIC", "Subjects",
	ETYPE_FILE, 
	"/r43/list.files/eric.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ERIC", "Single Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/eric.ssbj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.ssbj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"ERIC", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/eric.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"ERIC", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/eric.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ERIC", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/eric.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"ERIC", "Languages", 		
	ETYPE_FILE, 
	"/r43/list.files/eric.lang", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/eric.lang_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BINP", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/binp.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"BINP", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/binp.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BINP", "Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/binp.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.titl_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BINP", "Publishers", 		
	ETYPE_FILE, 
	"/r43/list.files/binp.publ", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.publ_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BINP", "Publisher Abbreviations",
	ETYPE_FILE, 
	"/r43/list.files/binp.pbab", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.pbab_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BINP", "Audiences",
	ETYPE_FILE, 
	"/r43/list.files/binp.audi", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/binp.audi_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BPUB", "Company Names", 		
	ETYPE_FILE, 
	"/r43/list.files/bpub.comp", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/bpub.comp_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BPUB", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/bpub.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/bpub.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"CMPY", "Company Names", 		
	ETYPE_FILE, 
	"/r43/list.files/cmpy.comp", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/cmpy.comp_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"CMPY", "Primary SIC", 		
	ETYPE_FILE, 
	"/r43/list.files/cmpy.psic", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/cmpy.psic_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"CMPY", "SIC Codes", 		
	ETYPE_FILE, 
	"/r43/list.files/cmpy.sicc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/cmpy.sicc_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"CMPY", "State Codes", 		
	ETYPE_FILE, 
	"/r43/list.files/cmpy.stat", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/cmpy.stat_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"LCAT", "Exact Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.exbj_hdr", NULL,	/* Header.  xxx special case. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"LCAT", "Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.titl_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Exact Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.exti_hdr", NULL,	/* Header.  xxx special case. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"LCAT", "Locations", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.locn", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.locn_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Instructor", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.prof", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.prof_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Course Name", 		
	ETYPE_FILE, 
	"/r43/list.files/lcat.cour", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.cour_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Call Numbers",
	ETYPE_FILE, 
	"/r43/list.files/lcat.call", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.call_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"LCAT", "Languages",
	ETYPE_FILE, 
	"/r43/list.files/lcat.lang", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/lcat.lang_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"MEDL", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/medl.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"MEDL", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/medl.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"MEDL", "Languages", 		
	ETYPE_FILE, 
	"/r43/list.files/medl.lang", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.lang_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"MEDL", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/medl.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"MEDL", "Single Subjects", 
	ETYPE_FILE, 
	"/r43/list.files/medl.ssbj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.ssbj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"MEDL", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/medl.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/medl.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME83", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/me83.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me83.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME83", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/me83.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me83.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME83", "Languages", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Languages"},

    {"ME83", "Subjects", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Subjects"},

    {"ME83", "Single Subjects", 
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Single Subjects"},

    {"ME83", "Publication Types", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Publication Types"},

    {"ME78", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/me78.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me78.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME78", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/me78.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me78.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME78", "Languages", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Languages"},

    {"ME78", "Subjects", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Subjects"},

    {"ME78", "Single Subjects", 
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Single Subjects"},

    {"ME78", "Publication Types", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Publication Types"},

    {"ME73", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/me73.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me73.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME73", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/me73.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me73.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME73", "Languages", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Languages"},

    {"ME73", "Subjects", 		
	ETYPE_REF, 
	"/r43/list.files/Headers/me73.joti_hdr", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Subjects"},

    {"ME73", "Single Subjects", 
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Single Subjects"},

    {"ME73", "Publication Types", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Publication Types"},

    {"ME66", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/me66.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me66.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME66", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/me66.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/me66.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"ME66", "Languages", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Languages"},

    {"ME66", "Subjects", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Subjects"},

    {"ME66", "Single Subjects", 
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"MEDL", "Single Subjects"},

    {"ME66", "Publication Types", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	"MEDL", "Publication Types"},

    {"GROL", "Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/grol.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/grol.titl_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"INSP", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"INSP", "Single Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.ssbj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.ssbj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"INSP", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"INSP", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"INSP", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"INSP", "Languages", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.lang", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.lang_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"INSP", "Treatment Codes", 		
	ETYPE_FILE, 
	"/r43/list.files/insp.trmt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/insp.trmt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"PSYC", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"PSYC", "Single Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.ssbj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.ssbj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"PSYC", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"PSYC", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"PSYC", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"PSYC", "Languages", 		
	ETYPE_FILE, 
	"/r43/list.files/psyc.lang", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/psyc.lang_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"HDBK", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/hdbk.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/hdbk.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"HDBK", "Tbl-Contents", 		
	ETYPE_FILE, 
	"/r43/list.files/hdbk.tblc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/hdbk.tblc_hdr", NULL,	/* Header. */
	-1, NULL, 1, 0, 
	NULL, NULL},

    {"HDBT", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/hdbt.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/hdbt.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"HDBT", "Tbl-Contents", 		
	ETYPE_FILE, 
	"/r43/list.files/hdbt.tblc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/hdbt.tblc_hdr", NULL,	/* Header. */
	-1, NULL, 1, 0, 
	NULL, NULL},

    {"GENC", "Courses", 		
	ETYPE_FILE, 
	"/r43/list.files/genc.cour", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/genc.cour_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"GENC", "Tbl-Contents", 		
	ETYPE_FILE, 
	"/r43/list.files/genc.tblc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/genc.tblc_hdr", NULL,	/* Header. */
	-1, NULL, 1, 0, 
	NULL, NULL},

    {"GCAT", "Languages", 		
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"LCAT", "Languages"},     /* use lcat language list */

    {"GCAT", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/gcat.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gcat.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"WCAT", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"WCAT", "Exact Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.exbj_hdr", NULL,	/* Header.  xxx special case. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"WCAT", "Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.titl_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Exact Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.titl", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.exti_hdr", NULL,	/* Header.  xxx special case. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"WCAT", "Locations", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.locn", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.locn_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Instructor", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.prof", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.prof_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Course Name", 		
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.cour", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.cour_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Call Numbers",
	ETYPE_FILE, 
	"/r43/list.files/New_files/wcat.call", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/wcat.call_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"WCAT", "Languages",
	ETYPE_REF, 
	"", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	"LCAT", "Languages"},     /* use lcat language list */

    {"TULP", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/tulp.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/tulp.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"TULP", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/tulp.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/tulp.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"TULP", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/tulp.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/tulp.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"BUSI", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"BUSI", "Publication Types", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"BUSI", "Authors", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"BUSI", "Journal Titles", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"BUSI", "Industry Categories", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.incd", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.incd_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BUSI", "SIC Codes", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.sicc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.sicc_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BUSI", "Company Names", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.comp", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.comp_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"BUSI", "Ticker Symbols", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.tick", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.tick_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"BUSI", "Illustrations", 		
	ETYPE_FILE, 
	"/r43/list.files/busi.ills", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/busi.ills_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"GPER", "Subjects", 		
	ETYPE_FILE, 
	"/r43/list.files/gper.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"GPER", "Publication Types",
	ETYPE_FILE, 
	"/r43/list.files/gper.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"GPER", "Authors",
	ETYPE_FILE, 
	"/r43/list.files/gper.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"GPER", "Journal Titles",
	ETYPE_FILE, 
	"/r43/list.files/gper.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"GPER", "Industry Categories",
	ETYPE_FILE, 
	"/r43/list.files/gper.incd", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.incd_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"GPER", "SIC Codes",
	ETYPE_FILE, 
	"/r43/list.files/gper.sicc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.sicc_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"GPER", "Company Names",
	ETYPE_FILE, 
	"/r43/list.files/gper.comp", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.comp_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"GPER", "Ticker Symbols",
	ETYPE_FILE, 
	"/r43/list.files/gper.tick", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.tick_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"GPER", "Illustrations",
	ETYPE_FILE, 
	"/r43/list.files/gper.ills", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/gper.ills_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"PAPR", "Subjects",
	ETYPE_FILE, 
	"/r43/list.files/papr.subj", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.subj_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"PAPR", "Publication Types",
	ETYPE_FILE, 
	"/r43/list.files/papr.pubt", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.pubt_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"PAPR", "Authors",
	ETYPE_FILE, 
	"/r43/list.files/papr.auth", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.auth_hdr", NULL,	/* Header. */
	-1, NULL, 0, 2, 
	NULL, NULL},

    {"PAPR", "Journal Titles",
	ETYPE_FILE, 
	"/r43/list.files/papr.joti", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.joti_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"PAPR", "SIC Codes",
	ETYPE_FILE, 
	"/r43/list.files/papr.sicc", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.sicc_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"PAPR", "Company Names",
	ETYPE_FILE, 
	"/r43/list.files/papr.comp", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.comp_hdr", NULL,	/* Header. */
	-1, NULL, 0, 1, 
	NULL, NULL},

    {"PAPR", "Ticker Symbols",
	ETYPE_FILE, 
	"/r43/list.files/papr.tick", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.tick_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {"PAPR", "Illustrations",
	ETYPE_FILE, 
	"/r43/list.files/papr.ills", NULL,	/* List. */
	NULL, NULL,	/* Index. */
	"/r43/list.files/Headers/papr.ills_hdr", NULL,	/* Header. */
	-1, NULL, 0, 0, 
	NULL, NULL},

    {NULL, NULL,
	0, 
	NULL, NULL,	/* List. */
	NULL, NULL,	/* Index. */
	NULL, NULL,	/* Header. */
	0, NULL, 0, 0, 
	NULL, NULL},
};



/*
 * Process command line options.
 * Some options call for non-server type of actions, in which case 
 * we execute them here and exit.
 */
static void
ProcessOptions (int argc, char *argv[])
{
    int			c;
    extern char		*optarg;
    extern int		opterr;


    ProgName = argv[0];
    opterr = 0;
    while ((c = getopt(argc, argv, "dn:p:f:r:l:LTEW")) != EOF) {
	switch (c) {
	case 'n':
	    ClientLimit = atoi (optarg);
	    break;

	case 'p':
	    ServerPort = atoi (optarg);
	    break;

	case 'd':
	    ++Debug;
	    break;

	case 'f':
	    LogSetDir (optarg);
	    break;

	case 'l':
	    if (*optarg == '-') {
		fprintf (stderr, "Bad log options for -l\n");
		++opterr;
	    }
	    else
		LogSetMask (optarg);
	    break;

	case 'r':
	    LogSetRate (atoi(optarg));
	    break;

	case 'T':
	    FListTest ();
	    exit (0);

	case 'L':
	    FListPrint ();
	    exit (0);

	case 'E':
	    FListExercise ();
	    exit (0);

	case 'W':
	    TestLogger ();
	    exit (0);
	    
	case '?': 
	default:
	    opterr++;
	    goto Usage;
	}
    }

 Usage:
    if (opterr) {
	fprintf (stderr, "usage: %s [-n n] [-d]  \n",
		 ProgName);
	fprintf (stderr, " -n n    - Each server process handles up to n clients.\n");
	fprintf (stderr, "           default is %d\n", DEF_CLIENT_LIMIT);
	fprintf (stderr, "           limit is %d - open list files (max of %d)\n", 
		 getdtablesize(), (FListCount () * FILES_PER_LIST));
	fprintf (stderr, " -p n    - Server listens on port n\n");
	fprintf (stderr, "           default is %d\n", SERV_TCP_PORT);
	fprintf (stderr, " -f dir  - Write log files in directory 'dir'\n");
	fprintf (stderr, " -l ops  - Specify what to log.  ops is comma seperated list\n");
	fprintf (stderr, "           of the following:  message,error,lookup,client,\n");
	fprintf (stderr, "           proc,all.\n");  
	fprintf (stderr, " -r p    - Sets the percentage of lookups that are logged.\n");
	fprintf (stderr, "           Default is 100%.\n");
	fprintf (stderr, " -d      - generate debugging output to stderr.\n");
	fprintf (stderr, "           more -d for more debugging.\n");
	fprintf (stderr, "           Also has effect of keeping server in forground.");
	fprintf (stderr, " -L	   - Generate a list of lists served\n");
	fprintf (stderr, " -T	   - Test list and names.\n");
	fprintf (stderr, " -E	   - Verify access to all list files.\n");
	fprintf (stderr, "options L and E do not start the server.\n");
	exit (1);
    }
}




/*
 * Clean up child processes.  One signal may indicate serverl process
 * termination.
 */
ReapChild (sig, code) 
    int sig;
    int code;

{
    pid_t       cpid;
    int         status;

    if (Debug && cpid > 0)
	fprintf (stderr, "%s: ReapChild entered\n", ProgName);

    do {
	cpid = waitpid (-1, &status, WNOHANG);
	if (Debug && cpid > 0)
		fprintf (stderr, "%s: Process %d terminates\n", 
			 ProgName, cpid);
    } while (cpid > 0);
}



/*
 * Close files.  Called in responce to a SIGHUP
 */
ReloadFiles ()

{
    if (Debug)
	fprintf (stderr, "%s:  Reloading files (SIGHUP)\n", ProgName);
    LogMsg (LOG_MSG, "Closing files (SIGHUP)");
    close_files();
}



/*
 * SIGINT handler.  Cleanup and exit.
 */
CleanUpAndExit ()

{
    if (Debug)
	fprintf (stderr, "%s:  Got SIGINT\n", ProgName);
    LogMsg (LOG_MSG, "Server exiting (by signal)");
    LogFlush ();
    LogClose ();
    close_files();
    exit (0);
}


/*
 * Restart the log files.
 */
RestartLogFiles ()
{
    LogMsg (LOG_MSG, "Closing log file generation (by signal)");
    LogNewGeneration ();
}


    

    


main(int argc, char *argv[])
  {
    int sockfd;
    struct sockaddr_in serv_addr;
    struct hostent *hp;
    char *hn;
    int  on, rslt = 0;


    /* Process options.  ProcessOptions may not return, see function. */
    ClientLimit = DEF_CLIENT_LIMIT;
    ServerPort = SERV_TCP_PORT;
    RequestingClientID = 0;
    Debug = 0;
    ProcessOptions (argc, argv);


    /* Generate first log message before moving to background. */
    if (LogOpen () != 0) {
	perror ("Failed to open log file");
	exit (1);
    }
    LogMsg (LOG_MSG, "Server started");
    LogOptions ();


    /* If no debugging then move process to background. */
    if (Debug >= 1) {
	fprintf (stderr, "Debugging set to %d, client limit %d, server port %d\n", 
		Debug, ClientLimit, ServerPort);
    }
    else {
	LogFlush ();
	daemon_init ();
    }



    /* Open the socket on which we listen for connections. */
    if (Debug >= 3) 
	fprintf (stderr, "%s:  Open server socket\n", ProgName);
    if ((sockfd=socket(AF_INET, SOCK_STREAM, 0)) < 0) { 
	LogPerror("brs_server: can't open stream socket");
        return 1;
    }
    bzero( (char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family              = AF_INET;
    serv_addr.sin_addr.s_addr         = htonl(INADDR_ANY);
    serv_addr.sin_port                = htons(ServerPort);

    on = 1;
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
      { LogPerror("brs_server: can't bind local address");
        return 2;
       }
    listen(sockfd, 10);


    /* Set signal handlers.   If in forground we send the log to
     * stderr so there is no need for a signal that restarts 
     * the log file.  But if in background we use SIGINT to
     * restart the logfiles. */
    signal(SIGPIPE, SIG_IGN);      /* ignore broken pipe situation */
    signal(SIGCHLD, ReapChild);
    signal(SIGHUP, ReloadFiles);
    signal(SIGTERM, CleanUpAndExit);
    if (Debug >= 1)
	    signal(SIGINT, CleanUpAndExit);
    else
	    signal(SIGINT, RestartLogFiles);

	

    /* Process requests.  (actually never returns.) */
    ProcessRequests (sockfd);


    /* Done, though we never reach here. */
    close_files();
}  /* end of main */



/*
 * Move to background by:
 *	Forking and exiting parent process.
 *	become session leader.
 *	chdir to root.
 *	clear umask.
 */
void
daemon_init (void)
{
    pid_t	pid;

    fflush (stdout);
    pid = fork (); 
    if (pid < 0) {
	LogPerror ("daemon_init:  fork()");
	exit (-1);
    }
    if (pid != 0) {
        printf ("List server started, pid = %d\n", pid);
	exit (0);	/* Parent exits. */
    }

    /* Server dissasociates itself with interactive environment. */
#if 0
    fclose (stdin);
    fclose (stdout);
    fclose (stderr);
#endif
    setsid ();
    chdir ("/");
    umask (0);
		
    return;
}



/*
 * Main server logic in which:
 *	Connections are accepted,
 *	Requests are serviced,
 *	and Auxiliarly processes are created.
 */
void
ProcessRequests (int sockfd)
{
    int		clilen;
    int		childCount;	/* Count of children spawned. */
    int		i;
    int		width;
    fd_set	rfds, efds;
    struct sockaddr_in	cli_addr;
    Client	*clienttab;	/* Client table. */
    Client	*clnt;		/* Entry in client table. */
    int		newsock;	/* newly accepted socket. */
    unsigned int nextClntID;	/* Next client ID to assign.*/
    int		clientCount;	/* Number of active clients. */
    int		firstProc;	/* TRUE for original process only.*/
    int		rc;
    int         on;
    int		cpid;


    if (Debug >= 2) 
	fprintf (stderr, "%s:  ProcessRequests entered\n", ProgName);

    
    /* Allocate a socket table.*/
    clienttab = (Client *) malloc (sizeof (Client) * ClientLimit);
    if (clienttab == NULL) {
	LogPerror ("malloc client table");
	exit (1);
    }
    for (i = 0, clnt = clienttab; i < ClientLimit; ++i, ++clnt) 
	clnt->socket = -1;
    clientCount = 0;
    firstProc = TRUE;
    ChildNo = 0;
    childCount = 0;
    nextClntID = 0;
    on = 1;


    /* Process all requests. */
    for ( ; ; ) { 
	
	/* Build FD mask. */
	width = sockfd;
	FD_ZERO (&rfds);
	for (i = 0, clnt = clienttab; i < ClientLimit; ++i, ++clnt) {
	    if (clnt->socket != -1) {
		FD_SET (clnt->socket, &rfds);
		if (clnt->socket > width) 
		    width = clnt->socket;
	    }
	}
	bcopy (&rfds, &efds, sizeof (fd_set));
	++width;		/* Width must be one more than max. */


	/* For first process also include the server port in the list
	 * of sockets we check for read ready. */
	if (firstProc)
	    FD_SET (sockfd, &rfds);


	/* Select until something to do. */
	if (Debug >= 4) 
	    fprintf (stderr, "  PR%d:  Calling select, width %d, clientCount %d\n",
		     ChildNo, width, clientCount);
	rc = select (width, &rfds, NULL, &efds, NULL);
	if (rc == -1) {
	    if (errno != EINTR)
	    	LogPerror ("select");
	    continue;
	}
	SelectTime = GetTickCount ();

	
	/* Ready to accept a connection? */
	if (firstProc && FD_ISSET (sockfd, &rfds))  {

	    /* Accept new connection. */
	    clilen = sizeof(cli_addr);
	    newsock = accept (sockfd, (struct sockaddr *) &cli_addr,
			&clilen);
	    if (newsock < 0) {
		LogPerror("brs_server: accept error");
		continue;
	    }
	    rc = setsockopt (newsock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
	    if (rc != 0) 
		LogPerror ("setsockopt");


	    /* Find empty client table entry and insert the new
	     * socket.  */
	    for (i = 0, clnt = clienttab; i < ClientLimit; ++i, ++clnt) 
		if (clnt->socket == EMPTY_SOCKET) break;
	    if (Debug >= 1) 
		assert (i < ClientLimit);
	    clnt->socket = newsock;
	    clnt->id = nextClntID++;
	    clnt->requests = 0;
	    ++clientCount;
	    if (Debug >= 3) 
		fprintf (stderr, "  PR%d:  Accept new client, sock %d, clientCount %d\n",
			ChildNo, clnt->socket, clientCount);

	    clnt->log_me = LogClientConnect (clnt->id, &cli_addr);

	    /* If we are at the client limit, fork a process to handle
	     * this passel of clients while the main process continues
	     * to listen for clients. */
	    if (clientCount == ClientLimit) {
		++childCount;
		LogPrepareToFork ();
		cpid = fork ();

		if (cpid == -1) {
		    /* Fork Failed! */
		    LogPerror ("fork ()  child server");
		    LogClientDiscon (clnt->id, "Fork Failed", clnt->requests);
		    --clientCount;
		    --childCount;
		    close (clnt->socket);
		    clnt->socket = EMPTY_SOCKET;
		}

		else if (cpid == 0) {
		    /* Child closes server socket and services clients. 
		     * close and re-open database files (otherwise file
		     * descriptors are shaired and bad things happen). */
		    LogClose ();
		    close (sockfd);
		    close_files();
		    firstProc = FALSE;
		    ChildNo = childCount;
		    LogThisProcExit = LogProcCreate (getpid ());
		    if (Debug >= 2)
	 		fprintf (stderr, "  PR%d:  Fork child %d (pid %d) to handle connections\n", 
			    ChildNo, childCount, cpid);
		}

		else {
		    /* Server closes client sockets. */
		    for (i = 0,clnt = clienttab; i < ClientLimit; ++i,++clnt) {
			if (clnt->socket != EMPTY_SOCKET) {
			    close (clnt->socket);
			    clnt->socket = EMPTY_SOCKET;
			}
		    }
		    clientCount = 0;
		}

		continue;	/* Go back to select. */
	    }
	}



	/* For each of the client sockets check if there is data. */
	for (i = 0, clnt = clienttab; i < ClientLimit; ++i, ++clnt) {
	    if (clnt->socket != EMPTY_SOCKET) {

		/* if set then request received, service it. */
		if (FD_ISSET (clnt->socket, &rfds)) {
		    if (Debug >= 4) 
			fprintf (stderr, "  PR%d:  Read from socket %d\n", 
			    ChildNo, clnt->socket);
		    /* Initialize statistics for this lookup. */
		    LStats.client_id = clnt->id;
		    LStats.reads = 0;
		    LStats.writes = 0;
		    LStats.prefix = NULL;
		    LStats.line_req = 0;
		    ++clnt->requests;
		    RequestingClientID = clnt->id;
		    rc = str_extract(clnt->socket);
		    RequestingClientID = -1;

		    if (rc != REQUEST_SUCCESS) {
			/* if the lookup was not a success we drop client. */
			--clnt->requests;
			FD_SET (clnt->socket, &efds);
			if (rc == REQUEST_ERROR) {
			    /* If an actual error occured, we want to log
			     * client discon and proc exit. */
			    LogThisProcExit = TRUE;
			    clnt->log_me = TRUE;
			}
		    }
		}

		/* if exception then client closed socket.  Close this
		 * end.  If we are a clone server (!firstProc) and there
		 * are no client connections left we exit here. */
		if (FD_ISSET (clnt->socket, &efds)) {
		    if (clnt->log_me) 
			    LogClientDiscon (clnt->id, "Disconect", 
					     clnt->requests);
		    close (clnt->socket);
		    clnt->socket = EMPTY_SOCKET;
		    --clientCount;
		    if (Debug >= 3) 
			fprintf (stderr, 
			    "  PR%d:  Close client connetion, sock %d, clientCount %d\n", 
			    ChildNo, clnt->socket, clientCount);
		    if (clientCount == 0 && !firstProc) {
			if (LogThisProcExit)
				LogProcExit (getpid (), "Zero clients");
			LogClose ();
			close_files ();
			if (Debug >= 2)
			    fprintf (stderr, "  PR%d:  Child %d exits\n", 
				ChildNo, ChildNo);
			exit (0);
		    }
		}
	    }
	}
    }
}
 



/*
 * Read request and send responce.  
 * returns 
 *	REQUEST_SUCCESS if there were no errors.
 * 	REQUEST_ERROR	if there was an error.
 *	REQUEST_CLIENT_GONE	if the client has closed the connection.
 */
int str_extract(int sockfd)
   {
     int i, n, rc,
         length,
         line_no,
         rtn_off_lines,
         req_number;
     
     struct timeval timeout;
     fd_set fds; 
     char parameters[LINEMAX],
          check_appls[100],
          db_name[10],
          list_name[30],
          prefix_ind[10],
          prefix_st[LINEMAX],
          lines[11],
          rtn_offs[11],
          rn[11],
          work_line[LINEMAX];
     Tick RequestTime;

     char *ptr_start;


     RequestTime = GetTickCount ();

     n = read(sockfd, parameters, LINEMAX);
     if (n == 0)
         return REQUEST_CLIENT_GONE;
     else if (n < 0) {
	      if (errno != ECONNRESET) 
                  LogPerror("brs_server: readline error");
              return REQUEST_ERROR;
             }
     ptr_start = &parameters[0];
     length = readline(ptr_start, check_appls, n);
     if (strncmp(check_appls, "willow", 6) == 0)
         ;                   /* check users */
     else {
	 if (strcmp(check_appls, "re_open") == 0) {
	      if (Debug)
		  fprintf (stderr, "%s:  Closing files (client request)\n", ProgName);
	      sprintf (work_line, "CID=%d	Closing files (re_open)",
		       RequestingClientID);
	      LogMsg (LOG_MSG, work_line);
	      close_files();
              return REQUEST_SUCCESS;
	  }
          else
              return REQUEST_SUCCESS;     /* unknown users use the same port */
     }

     ptr_start += length;
     n -= length;
     length = readline (ptr_start, rn, n);		/* Request number. */
     req_number = atoi (rn);
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, db_name, n);		/* Database Name. */
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, list_name, n);	/* List Name. */
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, prefix_ind, n);
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, prefix_st, n);
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, lines, n);
     line_no = atoi(lines);
     if (line_no <= 0)
         line_no = 1;
     ptr_start += length;
     n -= length;
     length = readline(ptr_start, rtn_offs, n);
     rtn_off_lines = atoi(rtn_offs);
     if (rtn_off_lines < 0)
         rtn_off_lines = 0;

     if (Debug >= 4) 
	 fprintf (stderr, "  se:  Got Request, req_number %d, lines %d\n",
		  req_number, line_no);

     
     LStats.db_name = db_name;
     LStats.list_name = list_name;


     list_req = NULL;
     file_check(db_name, list_name);

     if (list_req == NULL) {
	 sprintf(work_line, "Can't open  %s %s file!",
                             db_name, list_name);
	 return (ErrorReplay (sockfd, req_number, work_line));
     }
     rc = pull_data(sockfd, req_number, &prefix_ind[0], prefix_st, line_no, 
			rtn_off_lines);
     
     if (rc == REQUEST_SUCCESS) {
	 Tick 		now;
	 now = GetTickCount ();
	 LStats.service_time = now - SelectTime;
	 LStats.lookup_time = now - RequestTime;
	 LogLookup (&LStats);
     }
     return rc;
 } /* end of str_extract */




int readline(ptr, output_str, linemax)
    register char *ptr, *output_str;
    register int linemax;
   {
     int n, rc;
     char c;

     for (n = 1; n < linemax; n++)
        { *output_str = *ptr++;
          if (*output_str == '\n')
              break;
          output_str++;
         } /* end of for loop */
     *output_str = '\0';
     return n;
   }  /* end of readline */

/* extract line_no lines from a list after the prefix */
 
static int  fold, dict, len;
 
int
pull_data (int fd, int req_number, char *indicator, char *prefix_str,
                             int line_no, int rtn_off_lines)
  {
        extern char     *optarg;
        extern int      optind;
        register long   bot, mid, top;
        register int    c;
        struct stat     sb;
        char   entry[LINEMAX],
               copy[LINEMAX],
               prefix[LINEMAX],
               off_set_line[LINEMAX],
               index_char[4],
	       ebuf[80];

        int  j, n;
        long i,
             req_line_no,
             off_set,
             idx_off_set,
             approximate_line;


 
        list = list_req;
        if (fstat(fileno(list), &sb)) {
	    sprintf (ebuf, "CID=%d	brs_server: fstat", RequestingClientID);
	    LogPerror(ebuf);
	    return (ErrorReplay (fd, req_number, 
				 "Internal List Server error #1"));
	}
        rewind(list);

        if (*indicator == 'H') {   /* request header info. */
	    LStats.lookup_type = 'H';
	    sprintf(entry, "%08d\n%ld%c%s\n%c", req_number, 
		    file_lines, RS, hdr_rec_ptr, EOT);
            n = strlen(entry);
	    ++LStats.writes;
            if (write(fd, entry, n) != n) {
		sprintf (ebuf, "CID=%d	write (header)", RequestingClientID);
		LogPerror (ebuf);
		return REQUEST_ERROR;
	    }
            return REQUEST_SUCCESS;
	}
        else if ( req_tblc == 1 ) {
	    LStats.lookup_type = 'T';
	    approximate_line = 0;
            line_no = file_lines;
            off_set = 0;
	}
        else if (*indicator == 'N') {
	    req_line_no = atol(prefix_str);

	    LStats.lookup_type = 'N';
	    LStats.line_req  = req_line_no;

            approximate_line = req_line_no;
            if (req_line_no <= 0) {
		req_line_no = 0;
                approximate_line = 0;
	    }
            else
               if (req_line_no > (file_lines - 1)) {
		   req_line_no = (file_lines - 1);
                   approximate_line = req_line_no;
	       }
            idx_off_set = (req_line_no - rtn_off_lines) * 4;
            if (idx_off_set < 0)
                idx_off_set = 0;
            (void)fseek(list_idx, idx_off_set, 0);
            fscanf(list_idx, "%c%c%c%c", &index_char[0], &index_char[1],
                                         &index_char[2], &index_char[3]);
	    ++LStats.reads;
            off_set = index_char[0] * 16777216 + index_char[1] * 65536
                      + index_char[2] * 256 + index_char[3]; 
	}
        else if (strcmp(prefix_str, "zzzzz") == 0) {
	    /* request for last line */
	    LStats.lookup_type = 'P';
	    LStats.prefix = prefix_str;
	    approximate_line = file_lines - 1;
	    idx_off_set = (approximate_line - rtn_off_lines) * 4;
	    if (idx_off_set < 0)
		idx_off_set = 0;
	    (void)fseek(list_idx, idx_off_set, 0);
	    fscanf(list_idx, "%c%c%c%c", &index_char[0], &index_char[1],
		   &index_char[2], &index_char[3]);
	    ++LStats.reads;
	    off_set = index_char[0] * 16777216 + index_char[1] * 65536
		+ index_char[2] * 256 + index_char[3]; 
	}
        else {
	    LStats.lookup_type = 'P';
	    LStats.prefix = prefix_str;
	    strcpy(prefix, prefix_str);
	    if (blk_to_hyphen == 1  || blk_to_hyphen == 2) {
		for (j = 0; prefix[j] != '\0'; j++) {
		    if (prefix[j] == ' ')
			prefix[j] = '-';
		}
	    }
	    len = strlen(prefix);
	    if (blk_to_hyphen == 2)
	      to_lower(prefix, prefix, '-'); /* fold it to lower case */
	    else
              to_lower(prefix, prefix, ' ');
	    len = strlen(prefix);
	    if (len > LINEMAX - 1)
		prefix[LINEMAX - 1] = '\0';
	    for (bot = 0, top = sb.st_size; ; ) {
		mid = (top + bot) / 2;
		if (fseek(list, mid, 0) != 0)
		    LogPerror ("fseek (1):");
		
		for (++mid; (c = fgetc(list)) != EOF && c != '\n'; ++mid);
		++LStats.reads;
		
		if (get_a_record(entry) <= 0)
		    break;
	        if (blk_to_hyphen == 2)
	            to_lower(entry, copy, '-');
	        else
		    to_lower(entry, copy, ' ');
		if (strncmp(prefix, copy, len) <= 0) {
		    if (top <= mid)
			break;
		    top = mid;
		}
		else
		    bot = mid;
	    }
            fseek(list, bot, 0);
            while ( (off_set = ftell(list)) < top) {
		if (get_a_record(entry) <= 0)
		    break;
	        if (blk_to_hyphen == 2)
	            to_lower(entry, copy, '-');
	        else
		    to_lower(entry, copy, ' ');
		if (strncmp(copy, prefix, len) >= 0)
		    break;
	    }
            if (off_set >= sb.st_size) {
                /* return last line */
                approximate_line = file_lines - 1;
                idx_off_set = (approximate_line - rtn_off_lines) * 4;
                if (idx_off_set < 0)
                    idx_off_set = 0;
                (void)fseek(list_idx, idx_off_set, 0);
                fscanf(list_idx, "%c%c%c%c", &index_char[0], &index_char[1],
		       &index_char[2], &index_char[3]);
		++LStats.reads;
                off_set = index_char[0] * 16777216 + index_char[1] * 65536
		    + index_char[2] * 256 + index_char[3]; 
	    }
            else {
		/* find the line number */
                approximate_line = line_search(off_set);
                idx_off_set = (approximate_line - rtn_off_lines) * 4;
                if (idx_off_set < 0)
                    idx_off_set = 0;
                (void)fseek(list_idx, idx_off_set, 0);
                fscanf(list_idx, "%c%c%c%c", &index_char[0], &index_char[1],
		       &index_char[2], &index_char[3]);
		++LStats.reads;
                off_set = index_char[0] * 16777216 + index_char[1] * 65536
		    + index_char[2] * 256 + index_char[3]; 
	    }
	}
        sprintf(off_set_line, "%08d\n%08ld\n", req_number, approximate_line);
        
	if (Debug >= 5) {
	    fputs ("Send responce:\n", stderr);
	    fputs (off_set_line, stderr);
	}

	n = strlen (off_set_line);
	++LStats.writes;
        if (write(fd, off_set_line, n) != n) {
	    sprintf (ebuf, "CID=%d:  write (off_set_line)", RequestingClientID);
	    LogPerror (ebuf);
            return REQUEST_ERROR;
	 }
        (void)fseek(list, off_set, 0);


        /* if it is a large lookup, allow TCP to collect bytes. */
	if (line_no > 200) {
	    int on = 0;
	    setsockopt (fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
	}


        /* Transmit lines requested. */
        i = 0;
        while ((n = getline(entry)) > 0 && i < line_no) {
	    ++LStats.writes;
	    if (write(fd, entry, n) != n) {
	        sprintf (ebuf, "CID=%d:  write (entry)", RequestingClientID);
		LogPerror (ebuf);
                return REQUEST_ERROR;
	    }
            i++;
	}
	LStats.lines_sent = line_no;

	off_set_line[0] = EOT;
	++LStats.writes;
	n = write (fd, off_set_line, 1);
	if (n < 0) {
	    sprintf (ebuf, "CID=%d:  write (EOT)", RequestingClientID);
	    LogPerror (ebuf);
	    return REQUEST_ERROR;
        }


	/* Return socket to setting for small lookups. */
	if (line_no > 200) {
	    int on = 1;
	    setsockopt (fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
	}

	return REQUEST_SUCCESS;
  }
 
/*
 * get_a_record --
 *      get a record to compare
 */
int 
get_a_record(char *buffer) 
{ 
    char *ptr_loc;
    
    ++LStats.reads;
    if ((fgets(buffer, LINEMAX, list)) == NULL)
	return 0;
    else {
        if ((ptr_loc = strchr(buffer, RS)) != 0)
	    *ptr_loc = '\0';
	return strlen(buffer);
    }
}
 
/*
 * getline --
 *      get a line
 */
int 
getline(char *buffer)
{
    ++LStats.reads;
    if ((fgets(buffer, LINEMAX, list)) == NULL)
	return 0;
    else
	return strlen(buffer);
}


 
/*
 *   fold it to lower case
 *        
 */
static void
to_lower(entry, copy, kp_char)
    register char *entry, *copy, kp_char;
{   
    register int  cnt;
    register char c;
    
    c = *entry++;
    cnt = LINEMAX - 1;
    for ( ; (c != '\0') && cnt; cnt--) {
	if (isalnum(c))
	    *copy++ = isupper(c) ? tolower(c) : c;
	else
	    if (c == kp_char)
		*copy++ = c;
	c = *entry++;
    }
    *copy = '\0';
}


/* find the exact line number inside a file */
long 
line_search(long offset) 
{
    long bot, mid, top,
        idx_off_set,
        text_off_set;
    char idx_char[4];
    
    for (bot = 0, top = (file_lines - 1); ; ) {
        mid = (top + bot) / 2;
	idx_off_set = mid * 4;
	fseek(list_idx, idx_off_set, 0);
	fscanf(list_idx, "%c%c%c%c", &idx_char[0], &idx_char[1],
	       &idx_char[2], &idx_char[3]);
	++LStats.reads;
	text_off_set = idx_char[0] * 16777216 + idx_char[1] * 65536
	    + idx_char[2] * 256 + idx_char[3]; 
	
	if (text_off_set == offset)
	    break;
	else {
            if (bot >= top)
                break;
            else {
                if (text_off_set > offset)
                    top = mid - 1;
                else
                    bot = mid + 1;
	    }
	}
    }
    return mid;
}




/* close all the list files */
void 
close_files(void) 
{ 
    LList		*l;

    for (l = list_files; l->db_name != NULL; ++l) {
	if (l->fp != NULL) {
	    fclose (l->fp);
	    l->fp = NULL;
	}
	if (l->lip != NULL) {
	    fclose (l->lip);
	    l->lip = NULL;
	}
	if (l->hp != NULL) {
	    fclose (l->hp);
	    l->hp = NULL;
	}
	if (l->hdr != NULL) {
	    free (l->hdr);
	    l->hdr = NULL;
	}
    }
}


 
/* 
 * Asssign file pointer.
 * Look up entry for db_name, list_name.
 * Set the following global variables:
 *	req_tblc
 *	blk_to_hyphen
 *	list_req
 *	list_idx
 *	list_lines
 *	hdr_rec_ptr
 */
void 
file_check(char *db_name, char *list_name)
{
    LList		*l;


    req_tblc = 0;
    blk_to_hyphen = 0;
    list_req = NULL;
    list_idx = NULL;
    file_lines = 0;
    hdr_rec_ptr = NULL;
    

    /*
     * Look up the list.
     */
    l = FListFind (db_name, list_name);
    if (l == NULL) {
	char buf[128];
	sprintf (buf, "CID=%d	Failed to find list \"%s\", \"%s\"",
		 RequestingClientID, db_name, list_name);
	LogMsg (ERROR_MSG, buf);
	return;
    }


    /*
     * If the list files are not open, open them now.
     */
    if (l->fp == NULL) {
	if (FListOpen (l) != SUCCESS) 
	    return;
    }


    /*
     * Copy value to globals.
     */
    req_tblc = l->req_tblc;
    blk_to_hyphen = l->blk_to_hyphen;
    list_req = l->fp;
    list_idx = l->lip;
    file_lines = l->file_lines;
    hdr_rec_ptr = l->hdr;
    return;
}



/*
 * Lookup 'db_name', 'list_name' in table.
 *
 * Simple minded linear search.
 */
LList *
FListFind (char *db_name, char *list_name)
{
    LList		*l;
    ListAlias		*la;

 Search:
    for (la = list_alias; la->alias != NULL; ++la) {
	if (strcmp (list_name, la->alias) == 0) {
	    list_name = la->name;
	    break;
	}
    }


    for (l = list_files; l->db_name != NULL; ++l) {
	if (strcmp (db_name, l->db_name) == 0) {
	    if (strcmp (list_name, l->list_name) == 0) {

		if (l->etype == ETYPE_FILE) {
		    /* Matched a file entry. */
		    return (l);
		}
		else {
		    /* Matched a reference entry.  Start over with
		     * new db and list names. */
		    db_name = l->rf_db_name;
		    list_name = l->rf_list_name;
		    goto Search;
		}

	    }
	}
    }
    return (NULL);
}




/* 
 * Calculate total lines of a file.
 */
long 
line_cnt(FILE *file_ptr)
{
    long total_line;
    struct stat  fs;
    
    total_line = 0;
    if (fstat(fileno(file_ptr), &fs)) { 
	LogPerror("brs_server: fstat in line_cnt");
	return (-1);
    }
    rewind (file_ptr);
    total_line = fs.st_size / 4;
    return (total_line);
}         


/*
 * Open the files for list 'l'.
 */
int
FListOpen (LList *l)
{
#if MAXPATH > HDR_MAX
    char		buf[MAXPATH];
#else
    char		buf[HDR_MAX];
#endif
    char		*lidx_name;

    if (l == NULL || l->f_name == NULL) {	
	sprintf (buf, "CID=%d	FListOpen called with bad list pointer",
		 RequestingClientID);
	LogMsg (ERROR_MSG, buf);
	goto Fail0;
    }
    

    /*
     * Open list file proper. 
     */
    l->fp = fopen (l->f_name, "r");
    if (l->fp == NULL) {
	LogFOpenErr (RequestingClientID, l->db_name, l->list_name, l->f_name, errno);
	goto Fail0;
    }

    /*
     * Open the line index file and count the lines.
     */
    if (l->lidx_name == NULL) {
	strcpy (buf, l->f_name);
	strcat (buf, LINE_INDEX_EXT);
	lidx_name = buf;
    }
    else
	lidx_name = l->lidx_name;
    l->lip = fopen (lidx_name, "r");
    if (l->lip == NULL) {
	LogFOpenErr (RequestingClientID, l->db_name, l->list_name, lidx_name, 
		     errno);
	goto Fail1;
    }
    l->file_lines = line_cnt (l->lip);
    if (l->file_lines == -1) 
	goto Fail2;


    /*
     * Open header file and read header.
     */
    l->hp = fopen (l->hdr_name, "r");
    if (l->hp == NULL) {
	LogFOpenErr (RequestingClientID, l->db_name, l->list_name, l->hdr_name,
		     errno);
	goto Fail2;
    }
    if (fgets (buf, HDR_MAX, l->hp) == NULL) {
	LogHeaderReadErr (l->db_name, l->list_name, l->hdr_name, errno);
	goto Fail3;
    }
    fclose(l->hp);
    l->hp == NULL;
    l->hdr = malloc (strlen (buf) + 2);
    if (l->hdr == NULL) {
	sprintf (buf, 
"CID=%d	malloc(%d) failed, errno %d.  For header of list \"%s\", \"%s\"",
		RequestingClientID,  strlen (buf) + 2, errno, l->db_name, 
		l->list_name);
	LogMsg (ERROR_MSG, buf);
	goto Fail3;
    }
    strcpy (l->hdr, buf);

    return (SUCCESS);



	
    /*
     * Error exit.  Undo opens and allocs that may have been done.
     */
 Fail3:	if (l->hp != NULL)
             fclose (l->hp);
        l->hp = NULL;
 Fail2:	fclose (l->lip);	
        l->lip = NULL;    
        l->file_lines = 0;
 Fail1:	fclose (l->fp);  	
        l->fp = NULL;
 Fail0:	return (FAIL);
}




int
FListCount (void)
{
    int		i;
    LList	*l;

    i = 0;
    for (l = list_files; l->db_name != NULL; ++l)
	++i;
    return (i);
}

	



/*
 * These next routines dump the tables and does a couple checks to
 * make sure it looks OK.  The 'names' table contains the file name
 * extensions that are commonly used for each type of list.  FListTest
 * verifys that each list file name has the proper extension to insure
 * we've not put a file name in the wrong place.
 */
ListAlias names[] = {
	{"Subjects", "subj"},
	{"Single Subjects", "ssbj"},
	{"Publication Types", "pubt"},
	{"Audiences", "audi"},
	{"Authors", "auth"},
	{"Journal Titles", "joti"},
	{"Languages", "lang"},
	{"Titles", "titl"},
	{"Publishers", "publ"},
	{"Publisher Abbreviations", "pbab"},
	{"Exact Subjects", "exbj"},
	{"Exact Titles", "exti"},
	{"Locations", "locn"},
	{"Instructor", "prof"},
	{"Course Name", "cour"},
	{"Treatment Codes", "trmt"},
	{"Tbl-Contents", "tblc"},
	{"Courses", "cour"},
	{"Industry Categories", "incd"},
	{"Primary SIC", "psic"},
	{"SIC Codes", "sicc"},
	{"State Codes", "stat"},
	{"Company Names", "comp"},
	{"Ticker Symbols", "tick"},
	{"Illustrations", "ills"},
	{"Call Numbers", "call"},
	{NULL, NULL}
    };

char *req_tblc_str[] = {"", "req_tblc, "};
char *blk_to_hyphen_str[] = {"", "blk_to_hyphen_DH, ", "blk_to_hyphen_KH, "};

char *
FindExt (char * name) 
{
    ListAlias	*la;

    for (la = names; la->alias != NULL; ++la) {
	if (strcmp (la->alias, name) == 0)
	    return (la->name);
    }
    return (NULL);
}

void
FListTest (void)
{
    LList		*l;
    ListAlias		*la;
    char		*ext;

    /* Dump the alises. */
    printf ("List name aliases\n");
    for (la = list_alias; la->alias != NULL; ++la) 
	printf ("\"%s\" -> \"%s\"\n", la->alias, la->name);

    
    printf ("\n\nList files\n");
    for (l = list_files; l->db_name != NULL; ++l) {
	if (l->etype == ETYPE_FILE) {
	    printf ("\"%s\", \"%s\",  %s%s\n    list \"%s\",\n    header \"%s\"\n",
		    l->db_name, l->list_name,	
		    req_tblc_str[l->req_tblc],
		    blk_to_hyphen_str[l->blk_to_hyphen],
		    l->f_name, l->hdr_name);

	    /* Print and verify index file name. */
	    if (l->lidx_name != NULL) {
		printf ("    line index \"%s\"\n", l->lidx_name);
		if (strlen (l->lidx_name) == 0)
		    printf ("ERROR missing list file name.\n");
	    }

	    /* Verify not empty strings. */
	    if (strlen (l->f_name) == 0)
		printf ("ERROR missing list file name.\n");
	    if (strlen (l->hdr_name) == 0) 
		printf ("ERROR missing header file name.\n");

	    /* Find extension and veryify. */
	    if ((ext = FindExt (l->list_name)) != NULL) {
		if (strstr (l->f_name, ext) == NULL) 
		    printf ("ERROR wrong extension on list name\n");
		if (strstr (l->hdr_name, ext) == NULL) 
		    printf ("ERROR wrong extension on header name\n");
		if (l->lidx_name != NULL)
		    if (strstr (l->lidx_name, ext) == NULL)
			printf ("ERROR wrong extension on line index name\n");
	    }
	    else
		printf ("ERROR Unknown list name\n(perhaps you should add it)\n");

	    if (strstr (l->hdr_name, "hdr") == NULL )
		printf ("ERROR header file does not have \"hdr\" extension\n");
	}
	else {
	    printf ("\"%s\", \"%s\"  ->  \"%s\", \"%s\"\n",
		l->db_name, l->list_name, l->rf_db_name, l->rf_list_name);
	}
    }
}


void
FListPrint (void)
{
    LList		*l;


    for (l = list_files; l->db_name != NULL; ++l) {
	if (l->etype == ETYPE_FILE) 
	    printf ("\"%s\", \"%s\"\n", l->db_name, l->list_name);
	else 
	    printf ("\"%s\", \"%s\"  ->  \"%s\", \"%s\"\n",
		l->db_name, l->list_name, l->rf_db_name, l->rf_list_name);
    }
}




void
FListExercise (void)
{
    LList	*l;
    struct stat  fs;

    fprintf (stderr, "Opening all lists...\n");
    for (l = list_files; l->db_name != NULL; ++l) {
	if (l->etype == ETYPE_FILE) {
	    if (FListOpen (l) != SUCCESS) {
		fprintf (stderr, "ERROR opening files for list \"%s\", \"%s\"\n",
			 l->db_name, l->list_name);
	    }
	    if (fstat(fileno(l->fp), &fs)) { 
		fprintf (stderr, "ERROR stating file \"%s\"  ", l->f_name);
		perror  ("fstat");
	    }
	}
    }

    fprintf (stderr, "\n\nClosing all lists...\n");
    close_files ();
    return;
}



void 
TestLogger (void)
{
    struct sockaddr_in	cli_addr;
    LookupStats		ls;


    LogMsg (LOG_MSG, "Testing logger");

    errno = ENOENT;
    LogPerror ("errno set to ENOENT");

    cli_addr.sin_addr.s_addr = 0x11223344;
    LogClientConnect (33, &cli_addr);

    LogProcCreate (1999);

    LogFOpenErr (0, "LCAT", "Authors", "lcat.athors", ENOENT);
    
    LogHeaderReadErr ("LCAT", "Table", "lcat.table", 3);

    ls.client_id = 333;
    ls.service_time = 444;
    ls.lookup_time = 440;
    ls.db_name = "Data Base Name";
    ls.list_name = "List Name";
    ls.lookup_type = 'P';
    ls.line_req = 555;
    ls.prefix = "Prefi";
    ls.lines_sent = 66;
    ls.reads = 44;
    ls.writes = 55;
    LogLookup (&ls);

    ls.lookup_type = 'H';
    LogLookup (&ls);
    
    ls.lookup_type = 'T';
    LogLookup (&ls);

    LogClientDiscon (33, "Disconnect", 2);

    LogProcExit (1999, "Done");

    LogFlush ();

    LogClose ();
}




int
ErrorReplay (int sockfd, int req_number, char *msg)
{
    char	buf[256];
    int		rc;

    
    sprintf(buf, "%08d\n00000000\n%s\n%c",
	    req_number, msg, EOT);    if (Debug >= 4) {
	fputs ("send error responce:\n", stderr);
	fputs (buf, stderr);
    }
    rc = write(sockfd, buf, strlen(buf));
    if (rc < 0) 
	    return REQUEST_ERROR;
    return REQUEST_SUCCESS;
}
